import gzip
from abc import ABC, ABCMeta
from csv import DictWriter, DictReader
from typing import Union, TextIO, BinaryIO, List, Type, Dict

from filedatasource.datafile import DataReader, DataWriter, ReadMode, DataSourceError, Mode, DataFile


def open_file(fname: str, mode: Mode, encoding: str = 'utf-8'):
    """ Method to open a file depending on if it is compress with gzip or not.
    :param fname: The path to the file.
    :param mode: The open mode: Mode.APPEND, Mode.WRITE or Mode.READ.
    :param encoding: The file encoding.
    :return: A stream.
    """
    open_func = gzip.open if fname.lower().endswith('.gz') else open
    return open_func(fname, f'{mode.value}t', encoding=encoding, newline='')


class CsvData(DataFile, ABC):
    """ Abstract class for object that deals with CSV files. """
    __metaclass__ = ABCMeta

    @property
    def encoding(self) -> str:
        """
        :return: The file encoding.
        """
        return self.__encoding

    def __init__(self, file_or_io: [str, TextIO, BinaryIO], mode: Mode, encoding: str = 'utf-8'):
        """ Constructor.

        :param file_or_io: The file path or the file stream.
        :param mode: The open mode: Mode.APPEND, Mode.WRITE or Mode.READ.
        :param encoding: The file encoding.
        """
        DataFile.__init__(self, file_or_io)
        self.__encoding = encoding
        self._file = open_file(file_or_io, mode, self.encoding) if isinstance(file_or_io, str) else file_or_io

    def close(self) -> None:
        """ Close the file stream. """
        self._file.close()


class CsvWriter(CsvData, DataWriter):
    """
    A CSV writer to create a typical CSV file with head. It is very easy to use, only need to

    .. code-block:: python

        fieldnames = ['id', 'name', 'surname', 'address']
        with ConflictsWriter('data.csv', fieldnames) as writer:
            writer.write_row(id=1, name='John', surname='Smith', address='Oxford street')

    Also, if the file ends with .gz, the file will be compressed with gzip automatically.
    """
    @property
    def fieldnames(self) -> List[str]:
        """
        :return: The sequence of field names to use as CSV head.
        """
        return self._fieldnames

    def __init__(self, file_or_io: Union[str, TextIO, BinaryIO], fieldnames: Union[List[str], type, object] = None,
                 mode: Mode = Mode.WRITE, encoding: str = 'utf-8') -> None:
        """ Constructor of this CSV writer.

        :param file_or_io: The file path or an opened stream to use. If it is a file path and it ends in .gz, then
        a compressed file is created using gzip.
        :param fieldnames: The field names of this CSV.
        :param mode: The writing mode: Mode.APPEND or Mode.WRITE. By default Mode.WRITE.
        :param encoding: The encoding (it is only used if the parameter file_or_io is a file path).
        :raises ValueError: If mode is not Mode.WRITE or Mode.APPEND or if file_or_io is a file stream with
          write or append modes but this modes does not correspond to the mode parameter.
        """
        CsvData.__init__(self, file_or_io, mode, encoding)
        DataWriter.__init__(self, file_or_io, mode)
        self._fieldnames = self._parse_fieldnames(fieldnames)

        self._writer = DictWriter(self._file, fieldnames=self.fieldnames)
        if mode == Mode.WRITE:
            self._writer.writeheader()
            self.__num_row = 0
        else:
            self.__num_row = None

    def write_row(self, **row) -> None:
        """ Write a row.

        :param row: The dictionary or parameters to write.
        """
        self._writer.writerow(row)
        if self.__num_row is not None:
            self.__num_row += 1

    def __len__(self) -> int:
        """
        Calculate the number of rows in the file.
        :return: The number of rows in the data source.
        :raises DataFileError: If with this data source is not possible to calculate the number of rows.
          It is not possible to calculate if this comes from a file stream and it is opened as APPEND mode.
        """
        if self.__num_row is None:
            if self.file_name:
                with CsvReader(self.file_name, encoding=self.encoding) as reader:
                    self.__num_row = len(reader)
            else:
                raise DataSourceError(
                    f'The length of the data source cannot be computed if it is defined as a file stream, '
                    f'instead of a file path and this writer is opened in APPEND mode.')
        return self.__num_row


class CsvReader(CsvData, DataReader):
    """
    A CSV reader to read a typical CSV file with head. It is very easy to use. For example, if the file 'data.csv'
    contains:

    .. code-block:: text

        id,name,surname,address
        1,John,Smith,Oxford street
        ---

    It is only necessary to write:

    .. code-block:: python

        with ConflictsWriter('data.csv') as reader:
            for row in reader:
                print(row.id, row.name, row.surname, row.address)

    Also, if the file ends with .gz, the file will be read from a compressed file with gzip automatically.
    """
    @property
    def fieldnames(self) -> List[str]:
        """
        :return: The sequence of field names to use as CSV head.
        """
        return list(self._reader.fieldnames)

    def __init__(self, file_or_io: Union[str, TextIO, BinaryIO], mode: ReadMode = ReadMode.OBJECT,
                 encoding: str = 'utf-8', types: Union[List[Type], Dict[str, Type]] = None):
        """ Constructor of this CSV reader.

        :param file_or_io: The file path or an opened stream to use. If it is a file path and it ends in .gz, then
        the compressed file is read using gzip.
        :param encoding: The encoding (it is only used if the parameter file_or_io is a file path).
        :param mode: The default mode to read the rows. When the reader is iterated,
          it will return objects, dictionaries or lists depending on if the value of this parameter is ReadMode.OBJECT,
          ReadMode.DICTIONARY or ReadMode.LIST, respectively.
        :param types: The type of each field.
        :raises ValueError: If the read mode is not ReadMode.OBJECT, ReadMode.DICT or ReadMode.LIST.
        """
        if mode not in [ReadMode.OBJECT, ReadMode.DICT, ReadMode.LIST]:
            raise ValueError(f'The read mode only can be ReadMode.OBJECT, ReadMode.DICT or ReadMode.LIST, not {mode}.')
        super(CsvReader, self).__init__(file_or_io, Mode.READ, encoding)
        DataReader.__init__(self, file_or_io, mode)
        self._reader = DictReader(self._file)
        self.__length = None
        if types and not isinstance(types, List) and not isinstance(types, Dict):
            raise ValueError('If the parameter types is defined, it must be a list o dict.')
        self.__types = types if types else []

    def read_row(self) -> dict:
        """ Read a row of the CSV file.

        :return: An Python dictionary with fields that represents the information of the file. The name of these fields
        correspond with the name of the CSV head fields.
        """
        d = next(self._reader)
        if not self.__types:
            return d
        if isinstance(self.__types, List):
            return {k: (self.__types[i](v) if i < len(self.__types) else v) for i, (k, v) in enumerate(d.items())}

        return {k: (self.__types[k](v) if k in self.__types else v) for k, v in d.items()}

    def __len__(self) -> int:
        """ Calculate the number of rows. The first time you call this method it read the whole file once and
        it could do the algorithm a little bit slower.

        :return: The number of rows.
        :raises DataFileError: If with this data source is not possible to calculate the number of rows.
          It is not possible to calculate if this comes from a file stream.
        """
        if self.__length:
            return self.__length
        if self.file_name:
            with CsvReader(self.file_name, ReadMode.DICT, self.encoding) as reader:
                self.__length = sum(1 for _ in reader)
                return self.__length
        # return self.__length
        raise DataSourceError(f'The length of the data source cannot be computed if it is defined as a file stream '
                              f'instead of a file path.')



from collections import namedtuple
from inspect import getmembers, isroutine
from typing import Dict, Any, List


def to_identifier(s: str) -> str:
    """ Convert a string s into an valid identifier.
    :return: A str that can be used as a Python identifier.
    """
    s = 'n' + s if s and s[0].isdigit() else s
    return ''.join([c if c.isidentifier() or c.isdigit() else '_' for c in s])


def attributes2dict(obj: object) -> Dict[str, Any]:
    """ Convert an object with attributes into a dictionary. The attributes can be properties or public attributes.
    :param obj: The object to extract its attributes.
    :return: A dictionary with the name of the attribute and its value.
    """
    members = getmembers(obj, lambda member: not (isroutine(member)))
    return {att[0]: att[1] for att in members if not att[0].startswith('_')}


def attributes2list(obj: object) -> List[str]:
    """ Convert an object with attributes into a list of strings. The attributes can be properties or public
     attributes.
    :param obj: The object to extract its attributes.
    :return: A list with the name of attributes.
    """
    members = getmembers(obj, lambda member: not (isroutine(member)))
    return [att[0] for att in members if not att[0].startswith('_')]


def dict2obj(d: dict) -> object:
    """ Convert a dictionary in a Python object.
    :param d: The dictionary.
    :return: A Python object of type "Row" with the name of the keys as attributes and the values of the dictionary as
    the values of that attributes. If the dictionary has a key that is not a valid Python id, then, this method tries
    to convert it to an valid identifier replacing the incorrect characters by a _, or adding a _ at the start if
    the identifier starts with a number.
    """
    keys = [to_identifier(key) for key in d.keys()]
    return namedtuple('Row', keys)(*d.values()) if d else None


def dict2list(d: dict) -> List[Any]:
    """ Convert the values or a dictionary into a list.
    :param d: The dictionary.
    :return: A list with the values of that dictionary.
    """
    return [v for v in d.values()]


def dict_keys2list(d: dict) -> List[Any]:
    """ Convert the values or a dictionary into a list.
    :param d: The dictionary.
    :return: A list with the values of that dictionary.
    """
    return [v for v in d.keys()]



import os
import unittest
from typing import List

from tqdm import tqdm

from filedatasource import CsvWriter, CsvReader, ExcelWriter, ExcelReader, Mode, ReadMode, DataWriter, DataReader, \
    open_reader, open_writer, excel2list, excel2dict, csv2dict, csv2objects, objects2csv, dict2csv, list2csv, csv2list, \
    save, load, convert, load_lists, sheets
from filedatasource.csvfile import open_file
from filedatasource.datafile import DataSourceError

DATA_FILE = 'data.csv'
COMPRESSED_FILE = 'data.csv.gz'
EXCEL_FILE = 'data.xlsx'
XLS_FILE = 'data.xls'


def write_registers(writer) -> None:
    writer.write_row(a=1, b=2, c=3)
    writer.write_row(a=2, b=4, c=7)
    writer.write_row(a=3, b=6, c=15)
    writer.write_dict({'a': 4, 'b': 8, 'c': 31})


class Employee(object):
    @property
    def full_name(self) -> str:
        return self.name + ' ' + self.surname

    @property
    def name(self) -> str:
        return self.__name

    @property
    def surname(self) -> str:
        return self.__surname

    def __init__(self, name: str, surname: str):
        self.__name = name
        self.__surname = surname


class Example(object):
    def __init__(self, a: int, b: int, c: int):
        self.a, self.b, self.c = a, b, c


lists = [
    [1, 2, 3],
    [4, 5, 6],
    [7, 8, 9]
]
dicts = [
    {'a': 10, 'b': 11, 'c': 12},
    {'a': 13, 'b': 14, 'c': 15}
]
objects = [
    Example(16, 17, 18),
    Example(19, 20, 21),
    Example(22, 23, 24)
]


class TestWriter(CsvWriter):
    @property
    def fieldnames(self) -> List[str]:
        return ['a', 'b', 'c']


def list_to_int(lists: List[List[str]]) -> List[List[int]]:
    r = []
    for lst in lists:
        r.append([int(x) for x in lst])
    return r


class MyTestCase(unittest.TestCase):
    def test_write_csv_dict(self) -> None:
        with CsvWriter(DATA_FILE, fieldnames=['a', 'b', 'c']) as writer:
            self.assertEqual(len(writer), 0)
            write_registers(writer)
            self.assertEqual(len(writer), 4)
        self.assertTrue(os.path.exists(DATA_FILE))
        with CsvReader(DATA_FILE) as reader:
            for obj in tqdm(reader):
                pass
        self.assertEqual(obj.b, '8')
        os.remove(DATA_FILE)

    def test_write_csv_obj(self) -> None:
        with CsvWriter(COMPRESSED_FILE, fieldnames=Employee) as writer:
            writer.write(Employee('John', 'Smith'))
            writer.write(Employee('Maria', 'Ortega'))
        with CsvReader(COMPRESSED_FILE) as reader:
            obj = next(reader)
            self.assertEqual(obj.name, 'John')
            self.assertEqual(obj.surname, 'Smith')
            obj = next(reader)
            self.assertEqual(obj.name, 'Maria')
            self.assertEqual(obj.surname, 'Ortega')
            with self.assertRaises(StopIteration):
                reader.read()
        os.remove(COMPRESSED_FILE)

    def test_write_excel(self) -> None:
        with ExcelWriter(EXCEL_FILE, fieldnames=['a', 'b', 'c']) as writer:
            self.assertEqual(len(writer), 0)
            write_registers(writer)
            self.assertEqual(len(writer), 4)
        self.assertTrue(os.path.exists(EXCEL_FILE))
        with ExcelReader(EXCEL_FILE) as reader:
            for obj in tqdm(reader):
                pass
        self.assertEqual(obj.b, 8)
        os.remove(EXCEL_FILE)

    def test_import(self) -> None:
        with CsvWriter(DATA_FILE, fieldnames=['a', 'b', 'c']) as writer:
            self.assertEqual(len(writer), 0)
            write_registers(writer)
            self.assertEqual(len(writer), 4)
        self.assertTrue(os.path.exists(DATA_FILE))
        with CsvReader(DATA_FILE) as reader:
            with ExcelWriter(EXCEL_FILE, fieldnames=reader.fieldnames) as writer:
                writer.import_reader(reader)
        self.assertTrue(os.path.exists(EXCEL_FILE))
        os.remove(DATA_FILE)
        os.remove(EXCEL_FILE)
        with ExcelWriter(EXCEL_FILE, fieldnames=['a', 'b', 'c']) as writer:
            self.assertEqual(len(writer), 0)
            write_registers(writer)
            self.assertEqual(len(writer), 4)
        self.assertTrue(os.path.exists(EXCEL_FILE))
        with ExcelReader(EXCEL_FILE) as reader:
            with CsvWriter(DATA_FILE, fieldnames=reader.fieldnames) as writer:
                writer.import_reader(reader)
        self.assertTrue(os.path.exists(DATA_FILE))
        os.remove(DATA_FILE)
        os.remove(EXCEL_FILE)

    def test_dict_2_object(self) -> None:
        d = {'1': 1, 'G&S': 2}
        with CsvWriter(DATA_FILE, fieldnames=d) as writer:
            writer.write_dict(d)
            writer.write_list(list(d.values()))
        with CsvReader(DATA_FILE) as reader:
            for obj in reader:
                pass
        self.assertEqual(obj.n1, '1')
        self.assertEqual(obj.G_S, '2')
        os.remove(DATA_FILE)

    def test_lists(self) -> None:
        with CsvWriter(DATA_FILE, fieldnames=['a', 'b', 'c']) as writer:
            self.__write_lists_writer(writer)
        with CsvReader(DATA_FILE) as reader:
            self.check_lists_reader(reader)
        with CsvReader(DATA_FILE) as reader:
            self.check_dicts_reader(reader)
        with CsvReader(DATA_FILE) as reader:
            self.check_objects_reader(reader)
        os.remove(DATA_FILE)

    def check_lists_csv(self, lists: List[List[int]]) -> None:
        self.assertEqual(len(lists), 8)
        self.assertListEqual(lists[0], ['1', '2', '3'])
        self.assertListEqual(lists[7], ['22', '23', '24'])

    def check_lists_reader(self, reader: DataReader) -> None:
        lists = reader.read_lists()
        if isinstance(reader, CsvReader):
            self.check_lists_csv(lists)
        else:
            self.check_lists_excel(lists)

    def check_lists_excel(self, lists: List[List[int]]):
        self.assertEqual(len(lists), 8)
        self.assertListEqual(lists[0], [1, 2, 3])
        self.assertListEqual(lists[7], [22, 23, 24])

    def check_dicts_reader(self, reader: DataReader) -> None:
        dicts = reader.read_rows()
        if isinstance(reader, CsvReader):
            self.check_dicts_csv(dicts)
        else:
            self.check_dicts_excel(dicts)

    def check_dicts_csv(self, dicts: List[dict]) -> None:
        self.assertEqual(len(dicts), 8)
        self.assertDictEqual(dicts[0], {'a': '1', 'b': '2', 'c': '3'})
        self.assertDictEqual(dicts[7], {'a': '22', 'b': '23', 'c': '24'})

    def check_dicts_excel(self, dicts: List[dict]) -> None:
        self.assertEqual(len(dicts), 8)
        self.assertDictEqual(dicts[0], {'a': 1, 'b': 2, 'c': 3})
        self.assertDictEqual(dicts[7], {'a': 22, 'b': 23, 'c': 24})

    def check_objects_reader(self, reader: DataReader) -> None:
        objs = reader.read_objects()
        if isinstance(reader, CsvReader):
            self.check_objects_csv(objs)
        else:
            self.check_objects_excel(objs)

    def check_objects_csv(self, objs) -> None:
        self.assertEqual(len(objs), 8)
        self.assertEqual(objs[0].a, '1')
        self.assertEqual(objs[0].b, '2')
        self.assertEqual(objs[0].c, '3')
        self.assertEqual(objs[7].b, '23')
        self.assertEqual(objs[7].c, '24')

    def check_objects_excel(self, objs) -> None:
        self.assertEqual(len(objs), 8)
        self.assertEqual(objs[0].a, 1)
        self.assertEqual(objs[0].b, 2)
        self.assertEqual(objs[0].c, 3)
        self.assertEqual(objs[7].b, 23)
        self.assertEqual(objs[7].c, 24)

    @staticmethod
    def __write_lists_writer(writer: DataWriter) -> None:
        writer.write_lists(lists)
        writer.write_dicts(dicts)
        writer.write_objects(objects)

    def test_read_modes(self) -> None:
        with CsvWriter(DATA_FILE, fieldnames=['a', 'b', 'c']) as writer:
            self.__write_lists_writer(writer)
        with CsvReader(DATA_FILE, mode=ReadMode.DICT) as reader:
            for obj in reader:
                pass
        self.assertDictEqual(obj, {'a': '22', 'b': '23', 'c': '24'})
        with CsvReader(DATA_FILE, mode=ReadMode.LIST) as reader:
            for obj in reader:
                pass
        self.assertListEqual(obj, ['22', '23', '24'])
        with CsvReader(DATA_FILE, mode=ReadMode.OBJECT) as reader:
            for obj in reader:
                pass
        self.assertEqual(obj.b, '23')
        self.assertEqual(obj.c, '24')
        os.remove(DATA_FILE)

    def test_append(self) -> None:
        with TestWriter(COMPRESSED_FILE) as writer:
            self.assertEqual(len(writer), 0)
            writer.write_lists(lists)
            self.assertEqual(len(writer), 3)
        with TestWriter(COMPRESSED_FILE, mode=Mode.APPEND) as writer:
            self.assertEqual(len(writer), 3)
            writer.write_dicts(dicts)
            self.assertEqual(len(writer), 5)
        with CsvReader(COMPRESSED_FILE, mode=ReadMode.OBJECT) as reader:
            self.assertListEqual(reader.read_list(), ['1', '2', '3'])
            self.assertEqual(len(reader), 5)
            for obj in reader:
                pass
        self.assertEqual(obj.b, '14')
        self.assertEqual(obj.c, '15')

    def test_xls_xlsx(self) -> None:
        with ExcelWriter(EXCEL_FILE, fieldnames=['a', 'b', 'c']) as writer:
            self.__write_lists_writer(writer)
        with ExcelReader(EXCEL_FILE) as reader:
            self.check_lists_reader(reader)
        with ExcelWriter(XLS_FILE, fieldnames=['a', 'b', 'c']) as writer:
            self.__write_lists_writer(writer)
        with ExcelReader(XLS_FILE) as reader:
            self.check_lists_reader(reader)
        os.remove(EXCEL_FILE)
        os.remove(XLS_FILE)

    def test_builders(self) -> None:
        with open_writer(EXCEL_FILE, fieldnames=['a', 'b', 'c']) as writer:
            self.__write_lists_writer(writer)
        with ExcelReader(EXCEL_FILE) as reader:
            self.check_lists_reader(reader)
        self.check_lists_excel(excel2list(EXCEL_FILE))
        os.remove(EXCEL_FILE)
        with open_writer(XLS_FILE, fieldnames=['a', 'b', 'c']) as writer:
            self.assertEqual(len(writer), 0)
            self.__write_lists_writer(writer)
            self.assertEqual(len(writer), 8)
        with ExcelReader(XLS_FILE) as reader:
            self.check_dicts_reader(reader)
        self.check_dicts_excel(excel2dict(XLS_FILE))
        os.remove(XLS_FILE)
        with open_writer(DATA_FILE, fieldnames=['a', 'b', 'c']) as writer:
            self.__write_lists_writer(writer)
        with CsvReader(DATA_FILE) as reader:
            self.check_objects_reader(reader)
        self.check_objects_csv(csv2objects(DATA_FILE))
        os.remove(DATA_FILE)
        list2csv(COMPRESSED_FILE, lists, ['a', 'b', 'c'])
        self.assertListEqual(lists, list_to_int(csv2list(COMPRESSED_FILE)))
        dict2csv(COMPRESSED_FILE, dicts, ['a', 'b', 'c'])
        r = csv2dict(COMPRESSED_FILE)
        self.assertEqual(len(r), 2)
        self.assertDictEqual(r[0], {'a': '10', 'b': '11', 'c': '12'})
        self.assertDictEqual(r[1], {'a': '13', 'b': '14', 'c': '15'})
        objects2csv(COMPRESSED_FILE, objects, ['a', 'b', 'c'])
        with open_reader(COMPRESSED_FILE) as reader:
            obj = next(reader)
            self.assertEqual(obj.c, '18')
            obj = next(reader)
            self.assertEqual(obj.b, '20')
            obj = next(reader)
            self.assertEqual(obj.a, '22')
        with open_reader(COMPRESSED_FILE, mode=ReadMode.DICT) as reader:
            d = next(reader)
            self.assertEqual(d['c'], '18')
            d = next(reader)
            self.assertEqual(d['b'], '20')
            d = next(reader)
            self.assertEqual(d['a'], '22')
        with open_reader(COMPRESSED_FILE, mode=ReadMode.LIST) as reader:
            lst = next(reader)
            self.assertEqual(lst[2], '18')
            lst = next(reader)
            self.assertEqual(lst[1], '20')
            lst = next(reader)
            self.assertEqual(lst[0], '22')
        os.remove(COMPRESSED_FILE)

    def test_more_builders(self) -> None:
        save(EXCEL_FILE, objects)
        with open_reader(EXCEL_FILE) as reader:
            obj = next(reader)
            self.assertEqual(obj.c, 18)
            obj = next(reader)
            self.assertEqual(obj.b, 20)
            obj = next(reader)
            self.assertEqual(obj.a, 22)
        objs = load(EXCEL_FILE)
        self.assertEqual(objs[0].c, 18)
        self.assertEqual(objs[1].b, 20)
        self.assertEqual(objs[2].a, 22)
        dicts = load(EXCEL_FILE, mode=ReadMode.DICT)
        self.assertEqual(dicts[0]['c'], 18)
        self.assertEqual(dicts[1]['b'], 20)
        self.assertEqual(dicts[2]['a'], 22)
        os.remove(EXCEL_FILE)

    def test_sheets(self) -> None:
        with ExcelReader('Example.xls', sheet=0) as reader:
            self.check_clients_sheet(reader)
            self.assertEqual(len(reader), 2)
        with ExcelReader('Example.xls', sheet=1) as reader:
            self.check_suppliers_sheet(reader)
            self.assertEqual(len(reader), 3)
        with ExcelReader('Example.xls', sheet='Clients') as reader:
            self.check_clients_sheet(reader)
            self.assertEqual(len(reader), 2)
        with ExcelReader('Example.xls', sheet='Suppliers') as reader:
            self.check_suppliers_sheet(reader)
            self.assertEqual(len(reader), 3)
        with ExcelReader('Example.xlsx', sheet=0) as reader:
            self.check_clients_sheet(reader)
            self.assertEqual(len(reader), 2)
        with ExcelReader('Example.xlsx', sheet=1) as reader:
            self.check_suppliers_sheet(reader)
            self.assertEqual(len(reader), 3)
        with ExcelReader('Example.xlsx', sheet='Clients') as reader:
            self.check_clients_sheet(reader)
            self.assertEqual(len(reader), 2)
        with ExcelReader('Example.xlsx', sheet='Suppliers') as reader:
            self.check_suppliers_sheet(reader)
            self.assertEqual(len(reader), 3)

    def check_clients_sheet(self, reader: DataReader) -> None:
        row = next(reader)
        self.assertEqual(row.Name, 'John')
        self.assertEqual(row.Surname, 'Smith')
        self.assertEqual(row.Email, 'johnsmith@example.com')
        self.assertEqual(row.Phone, 72123423)
        row = next(reader)
        self.assertEqual(row.Name, 'Maria')
        self.assertEqual(row.Surname, 'Ortega')
        self.assertEqual(row.Email, 'mariaortega@example.com')
        self.assertEqual(row.Phone, 72112234)

    def check_suppliers_sheet(self, reader: DataReader) -> None:
        row = next(reader)
        self.assertEqual(row.Name, 'Bill')
        self.assertEqual(row.Surname, 'Gates')
        self.assertEqual(row.Company, 'Microsoft')
        self.assertEqual(row.Email, 'billgates@microsoft.com')
        self.assertEqual(row.Phone, 787877878)
        self.assertEqual(row.Charge, 'Adviser')
        row = next(reader)
        self.assertEqual(row.Name, 'Jeff')
        self.assertEqual(row.Surname, 'Bezos')
        self.assertEqual(row.Company, 'Amazon')
        self.assertEqual(row.Email, 'jeffbezos@amazon.com')
        self.assertEqual(row.Phone, 799998789)
        self.assertEqual(row.Charge, 'CEO')
        row = next(reader)
        self.assertEqual(row.Name, 'Elon')
        self.assertEqual(row.Surname, 'Musk')
        self.assertEqual(row.Company, 'Tesla')
        self.assertEqual(row.Email, 'elonmask@tesla.com')
        self.assertEqual(row.Phone, 740023028)
        self.assertEqual(row.Charge, 'Engineering')
        with self.assertRaises(StopIteration):
            next(reader)

    def test_strems(self) -> None:
        with open('data.csv', 'wt') as file:
            writer = CsvWriter(file, ['a', 'b', 'c'])
            self.assertEqual(len(writer), 0)
            writer.write_lists(lists)
            self.assertEqual(len(writer), 3)
            writer.write_dicts(dicts)
            self.assertEqual(len(writer), 5)
            writer.write_objects(objects)
            self.assertEqual(len(writer), 8)
        with open('data.csv', 'rt') as file:
            with self.assertRaisesRegex(DataSourceError, 'The length of the data source cannot be computed if it is '
                                                       'defined as a file stream instead of a file path.'):
                len(CsvReader(file))

        with open(DATA_FILE, 'at') as file:
            with self.assertRaisesRegex(ValueError, r'The reader is in mode Mode.WRITE but the file stream is in not '
                                                    r'in write mode \("at"\).'):
                CsvWriter(file, ['a', 'b', 'c'])
        with open_file(DATA_FILE, Mode.WRITE) as file:
            with self.assertRaisesRegex(ValueError, r'The reader is in mode Mode.APPEND but the file stream is in not '
                                                    r'in append mode \("wt"\).'):
                CsvWriter(file, ['a', 'b', 'c'], Mode.APPEND)
        with open_file(DATA_FILE, Mode.APPEND) as file:
            with self.assertRaisesRegex(DataSourceError,
                                        'The length of the data source cannot be computed if it is defined as a file '
                                        'stream, instead of a file path and this writer is opened in APPEND mode.'):
                len(CsvWriter(file, ['a', 'b', 'c'], Mode.APPEND))
        os.remove(DATA_FILE)

    def test_convert(self) -> None:
        write_registers(open_writer(COMPRESSED_FILE, fieldnames=['a', 'b', 'c']))
        convert(COMPRESSED_FILE, XLS_FILE)
        with open_reader(COMPRESSED_FILE, ReadMode.DICT) as reader1:
            with open_reader(XLS_FILE, ReadMode.DICT) as reader2:
                for dict1 in reader1:
                    dict2 = next(reader2)
                    self.assertDictEqual(dict1, dict2)
                self.assertEqual(len(reader1), len(reader2))
        os.remove(COMPRESSED_FILE)

    def test_excel_append(self) -> None:
        with ExcelWriter(XLS_FILE, fieldnames=['a', 'b', 'c']) as writer:
            writer.write_dicts(dicts)
            self.assertEqual(len(writer), 2)
        with ExcelWriter(XLS_FILE, mode=Mode.APPEND) as writer:
            writer.write_lists(lists)
            writer.write_objects(objects)
            self.assertEqual(len(writer), 8)
        with ExcelReader(XLS_FILE) as reader:
            self.assertDictEqual(reader.read_row(), {'a': 10.0, 'b': 11.0, 'c': 12.0})
            self.assertDictEqual(reader.read_row(), {'a': 13.0, 'b': 14.0, 'c': 15.0})
            self.assertListEqual(reader.read_list(), [1, 2, 3])
            self.assertListEqual(reader.read_list(), [4, 5, 6])
            self.assertListEqual(reader.read_list(), [7, 8, 9])
            self.assertListEqual(reader.read_list(), [16, 17, 18])
            self.assertListEqual(reader.read_list(), [19, 20, 21])
            self.assertListEqual(reader.read_list(), [22, 23, 24])
            self.assertEqual(len(reader), 8)
        os.remove(XLS_FILE)
        with ExcelWriter(EXCEL_FILE, fieldnames=['a', 'b', 'c']) as writer:
            writer.write_dicts(dicts)
            self.assertEqual(len(writer), 2)
        with ExcelWriter(EXCEL_FILE, mode=Mode.APPEND) as writer:
            writer.write_lists(lists)
            writer.write_objects(objects)
            self.assertEqual(len(writer), 8)
        with ExcelReader(EXCEL_FILE) as reader:
            self.assertDictEqual(reader.read_row(), {'a': 10.0, 'b': 11.0, 'c': 12.0})
            self.assertDictEqual(reader.read_row(), {'a': 13.0, 'b': 14.0, 'c': 15.0})
            self.assertListEqual(reader.read_list(), [1, 2, 3])
            self.assertListEqual(reader.read_list(), [4, 5, 6])
            self.assertListEqual(reader.read_list(), [7, 8, 9])
            self.assertListEqual(reader.read_list(), [16, 17, 18])
            self.assertListEqual(reader.read_list(), [19, 20, 21])
            self.assertListEqual(reader.read_list(), [22, 23, 24])
            self.assertEqual(len(reader), 8)
        os.remove(EXCEL_FILE)

    def test_odd_actions(self):
        # Checking writing a list is sorter than the field names
        with open_writer(COMPRESSED_FILE, fieldnames=['a', 'b', 'c']) as writer:
            writer.write_list([18324])
        lists = load_lists(COMPRESSED_FILE)
        self.assertEqual(len(lists), 1)
        self.assertListEqual(lists[0], ['18324', '', ''])
        with open_writer(DATA_FILE, fieldnames=['a', 'b', 'c']) as writer:
            writer.write_list([18324])
        lists = load_lists(DATA_FILE)
        self.assertEqual(len(lists), 1)
        self.assertListEqual(lists[0], ['18324', '', ''])
        with open_writer(EXCEL_FILE, fieldnames=['a', 'b', 'c']) as writer:
            writer.write_list([18324])
        lists = load_lists(EXCEL_FILE)
        self.assertEqual(len(lists), 1)
        self.assertListEqual(lists[0], [18324, None, None])
        with open_writer(XLS_FILE, fieldnames=['a', 'b', 'c']) as writer:
            writer.write_list([18324])
        lists = load_lists(XLS_FILE)
        self.assertEqual(len(lists), 1)
        self.assertListEqual(lists[0], [18324, '', ''])
        os.remove(DATA_FILE)
        os.remove(COMPRESSED_FILE)
        os.remove(EXCEL_FILE)
        os.remove(XLS_FILE)

    def test_csv_types(self) -> None:
        with CsvWriter(DATA_FILE, fieldnames=['a', 'b', 'c', 'd']) as writer:
            writer.write(['One', 2, 3.0, True])
        with CsvReader(DATA_FILE, types=[str, int, float, bool]) as reader:
            self.assertListEqual(reader.read_list(), ['One', 2, 3.0, True])
        with CsvReader(DATA_FILE, types={'a': str, 'b': int, 'c': float, 'd': bool}) as reader:
            self.assertListEqual(reader.read_list(), ['One', 2, 3.0, True])
        with CsvReader(DATA_FILE, types=[str, int, float]) as reader:
            self.assertListEqual(reader.read_list(), ['One', 2, 3.0, 'True'])
        l = csv2list(DATA_FILE, types=[str, int, float])
        self.assertListEqual(l[0], ['One', 2, 3.0, 'True'])
        d = csv2dict(DATA_FILE, types=[str, int, float])
        self.assertDictEqual(d[0], {'a': 'One', 'b': 2, 'c': 3.0, 'd': 'True'})
        o = csv2objects(DATA_FILE, types=[str, int, float])
        self.assertEqual(o[0].c, 3.0)
        os.remove(DATA_FILE)

    def test_sheet_names(self) -> None:
        self.assertListEqual(sheets('Example.xls'), ['Clients', 'Suppliers'])
        self.assertListEqual(sheets('Example.xlsx'), ['Clients', 'Suppliers'])

    def test_errors(self) -> None:
        with CsvWriter(DATA_FILE, fieldnames=['a', 'b', 'c', 'd']) as writer:
            writer.write(['One', 2, 3.0, True])
        with CsvReader(DATA_FILE, types=[int, int, float]) as reader:
            with self.assertRaisesRegex(ValueError, r'invalid literal for int\(\) with base 10: \'One\''):
                self.assertListEqual(reader.read_list(), ['One', 2, 3.0, 'True'])
        os.remove(DATA_FILE)


if __name__ == '__main__':
    unittest.main()



from abc import ABC, ABCMeta, abstractmethod
from enum import Enum, unique, auto
from typing import List, Union, TextIO, BinaryIO, Any, Dict, Sequence, Callable

from filedatasource.utils import dict2obj, attributes2list, attributes2dict, dict2list, dict_keys2list


@unique
class Mode(Enum):
    """ The file open mode. """
    APPEND = 'a'  # To append at the end of the file.
    WRITE = 'w'  # To start a new file, this will destroy the previous one with the same name.
    READ = 'r'  # To read the file.


@unique
class ReadMode(Enum):
    """ Modes which DataReader will return the rows. """
    OBJECT = auto()  # Return each row as object with attributes
    DICT = auto()  # Return each row as dictionary
    LIST = auto()  # Return each row as list of values


class DataSourceError(Exception):
    pass


class DataFile(ABC):
    """
    Abstract class for object that deals with data files.
    """
    __metaclass__ = ABCMeta

    @property
    @abstractmethod
    def fieldnames(self) -> List[str]:
        """ :return: The sequence of field names to use as CSV head. """
        pass

    @property
    def file_name(self) -> str:
        """ :return: The file name. """
        return self._fname

    def __init__(self, file_or_io: Union[str, TextIO, BinaryIO]) -> None:
        """ Constructor.
        :param file_or_io: The file path to the Data file.
        """
        self._fname = file_or_io if isinstance(file_or_io, str) else None
        self._file_stream = file_or_io if not isinstance(file_or_io, str) else None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.close()

    @abstractmethod
    def close(self) -> None:
        """ This method is called when finishes with this object, usually to close the file. """
        pass


class DataWriter(DataFile, ABC):
    __metaclass__ = ABCMeta

    @property
    def mode(self) -> Mode:
        return self.__mode

    def __init__(self, file_or_io: Union[str, TextIO, BinaryIO], mode: Mode):
        super(DataWriter, self).__init__(file_or_io)
        self.__check_mode(mode)
        self.__mode = mode

    def __check_mode(self, mode):
        if mode not in [Mode.WRITE, Mode.APPEND]:
            raise ValueError(f'The {type(self).__name__} only allows modes {Mode.WRITE} or {Mode.APPEND}, not {mode}')
        f_mode = self._file_stream.mode.strip() if self._file_stream and hasattr(self._file_stream, 'mode') else None
        if f_mode and mode == Mode.WRITE and f_mode[0] != 'w':
            raise ValueError(f'The reader is in mode {mode} but the file stream is in not in write mode ("{f_mode}").')
        if f_mode and mode == Mode.APPEND and f_mode[0] != 'a':
            raise ValueError(f'The reader is in mode {mode} but the file stream is in not in append mode ("{f_mode}").')

    @staticmethod
    def _parse_fieldnames(fieldnames: Union[List[str], Dict, object]) -> List[str]:
        """ Convert the fieldnames if their are defined as a dict or an object with attributes or properties in
        a list of fieldnames without values.
        If this argument is already a list, then return it without any modification.

        :param fieldnames: A list, dictionary or object.
        :return: A list of strings that represent the fieldnames.
        """
        if isinstance(fieldnames, List):
            return fieldnames
        if isinstance(fieldnames, dict):
            return dict_keys2list(fieldnames)
        if isinstance(fieldnames, object):
            return attributes2list(fieldnames)
        return []

    def write(self, o: Union[List, dict, object]) -> None:
        """ Write a row.

        :param o: A list, a dictionary or an object with the values of the row to store.
        """
        if isinstance(o, List):
            self.write_list(o)
        elif isinstance(o, dict):
            self.write_dict(o)
        else:
            self.write_object(o)

    @abstractmethod
    def write_row(self, **row) -> None:
        """ Write a row.

        :param row: The dictionary or parameters to write.
        """
        pass

    def write_dict(self, row: dict) -> None:
        """ Write a dictionary as a row.

        :param row: A dictionary with the fieldnames as a key and the row data as the dictionary values.
        """
        self.write_row(**row)

    def write_dicts(self, rows: Sequence[dict]) -> None:
        """ Write a list of dictionaries.

        :param rows:  The list of dictionaries. Each dictionary has to contain as keys the fieldnames and as value
        the row data to store.
        """
        for row in rows:
            self.write_dict(row)

    def write_object(self, o: object) -> None:
        """ Write an objects.

        :param o: The objects to write with public attributes or properties.
        """
        self.write_row(**attributes2dict(o))

    def write_objects(self, objects: Sequence[object]) -> None:
        """ Write a sequence of objects.

        :param objects: The sequence of objects to write with public attributes or properties.
        """
        for o in objects:
            self.write(o)

    def write_list(self, lst: list) -> None:
        """ Write a list of values.

        :param lst: The list of values. It is going to store in the same order than the fieldnames.
        """
        self.write_dict({field: lst[i] for i, field in enumerate(self.fieldnames) if i < len(lst)})

    def write_lists(self, lists: Sequence[list]) -> None:
        """ Write a sequences of lists as a sequence of rows

        :param lists: The sequence of rows as lists.
        """
        for lst in lists:
            self.write_list(lst)

    def import_reader(self, reader: 'DataReader') -> None:
        """ Import the content of a reader in this writer.

        :param reader: The reader to import.
        """
        for obj in reader:
            self.write(obj)


class DataReader(DataFile, ABC):
    """ A data reader to read very easy a file with data, usually, the typical CSV or Excel files. """
    __metaclass__ = ABCMeta

    def __init__(self, file_or_io: Union[str, TextIO, BinaryIO], mode: ReadMode = ReadMode.OBJECT) -> None:
        """ Constructor.

        :param file_or_io: The file or the IO stream to read.
        :param mode: The default mode to read the rows. When the reader is iterated,
        it will return objects, dictionaries or lists depending on if the value of this parameter is ReadMode.OBJECT,
        ReadMode.DICTIONARY or ReadMode.LIST, respectively.
        """
        super(DataReader, self).__init__(file_or_io)
        if mode not in [elem for elem in ReadMode]:
            ValueError(f'The mode argument must be ReadMode.OBJECT, ReadMode.DICT or ReadMode.LIST instead of {mode}.')
        self.__mode = mode

    def __iter__(self):
        return self

    def __next__(self):
        return self.read()

    def read(self) -> object:
        """ Read a row of the CSV file.

        :return: An Python object with fields that represents the information of the file. The name of these fields
        correspond with the name of the CSV head fields.
        """
        if self.__mode == ReadMode.DICT:
            return self.read_row()
        elif self.__mode == ReadMode.OBJECT:
            return self.read_object()
        return self.read_list()

    @abstractmethod
    def read_row(self) -> dict:
        """ Read a row of the file as a dict.

        :return: It must to return a dictionary where the keys are the fieldnames, and their values the row values.
        """
        pass

    def read_rows(self) -> List[dict]:
        """ Read the whole file and return a list of dictionaries with each one of the file rows.

        :return: A list of dictionaries that represents the list of file rows.
        """
        return self.__read_lists(self.read_row)

    def read_list(self) -> list:
        """ Read a row of the CSV file as a list.

        :return: A list of values with the data of each column of the file row.
        """
        return dict2list(self.read_row())

    def read_lists(self) -> List[list]:
        """ Read the whole file and return a list of lists with each one of the file rows.

        :return: A list of lists that represents the list of file rows.
        """
        return self.__read_lists(self.read_list)

    def read_object(self) -> object:
        """ Read a row of the CSV file as a Python object.

        :return: A list of object with the fieldnames as object attributes and of each column value as their values.
        If there is any character that cannot be represented as a Python identifier, it is changed by underscore '_'.
        If the name of the column starts with number, then a 'n' character is added.
        """
        return dict2obj(self.read_row())

    def read_objects(self) -> List[object]:
        """ Read the whole file and return a list of Python objects with each one of the file rows.

        :return: A list of objects that represents the list of file rows. See read_object() method for further
        information.
        """
        return self.__read_lists(self.read_object)

    @staticmethod
    def __read_lists(func: Callable) -> List[Any]:
        """ Call the function until it returns a StopIteration exception. Then, this method returns a list with
        the results of the each function call.

        :param func: The function to call.
        :return: A list of results.
        """
        lst = []
        try:
            while True:
                result = func()
                lst.append(result)
        except StopIteration:
            return lst



from .datafile import DataWriter, DataReader, ReadMode, Mode
from .csvfile import CsvReader, CsvWriter
from .excel import ExcelReader, ExcelWriter, sheets
from .builder import open_reader, open_writer, list2csv, dict2csv, objects2csv, list2excel, \
    dict2excel, objects2excel, csv2list, csv2dict, csv2objects, excel2list, \
    excel2dict, excel2objects, load, load_objs, load_dicts, load_lists, save, save_objs, save_dicts, save_lists, convert



from typing import Union, List, TextIO, BinaryIO, Any, Dict, Type

from filedatasource import CsvReader, ExcelReader, CsvWriter, ExcelWriter, Mode, ReadMode, DataWriter, DataReader
from filedatasource.utils import attributes2list, dict_keys2list


def open_reader(fname: str, mode: ReadMode = ReadMode.OBJECT) -> DataReader:
    """ Create a CsvReader or a ExcelReader with the parameters by default only from the file path and the extension
    of the file name. If it ends with .csv.gz or .gz, then a CSV file is created, however, if the extension
    is .xls or .xlsx, then an Excel file is read.

    :param fname: The path to the Excel or CSV file.
    :param mode: The default mode to read the rows. When the reader is iterated,
        it will return objects, dictionaries or lists depending on if the value of this parameter is ReadMode.OBJECT,
        ReadMode.DICTIONARY or ReadMode.LIST, respectively.
    :return: A CsvReader or a ExcelReader depending on the file extension.
    :raises ValueError: If the file name is not a CSV (compressed or not) or Excel (XLSX, XLS) file.
    """
    if fname.lower().endswith('.csv') or fname.lower().endswith('.csv.gz'):
        return CsvReader(fname, mode=mode)
    if fname.lower().endswith('.xls') or fname.lower().endswith('.xlsx'):
        return ExcelReader(fname, mode=mode)
    raise ValueError(f'The file name {fname} has to finish in .csv, .csv.gz, .xls, or .xlsx to use this function')


def open_writer(fname: str, fieldnames: List[str]) -> DataWriter:
    """ Create a CsvWriter or a ExcelWriter with the parameters by default only from the file path with a given
     extension, and the fieldnames. If the file extension ends with .csv.gz or .gz, then a CSV file is created,
     however, if the extension is .xls or .xlsx, then an Excel file is created.

    :param fname: The path to the Excel or CSV file.
    :param fieldnames: The name of the fields.
    :return: A CsvWriter or a ExcelWriter depending on the file extension.
    :raises ValueError: If the file name is not a CSV (compressed or not) or Excel (XLSX, XLS) file.
    """
    if fname.lower().endswith('.csv') or fname.lower().endswith('.csv.gz'):
        return CsvWriter(fname, fieldnames=fieldnames)
    if fname.lower().endswith('.xls') or fname.lower().endswith('.xlsx'):
        return ExcelWriter(fname, fieldnames=fieldnames)
    raise ValueError(f'The file name {fname} has to finish in .csv, .csv.gz, .xls, or .xlsx to use this function')


def list2csv(file_or_io: Union[str, TextIO, BinaryIO], data: List[List[Any]],
             fieldnames: Union[List[str], type, object] = None,
             mode: Mode = Mode.WRITE, encoding: str = 'utf-8') -> None:
    """ Write a sequences of lists as a sequence of rows in a CSV file.

    :param file_or_io: The file path or an opened stream to use. If it is a file path and it ends in .gz, then
    :param data: The sequence of rows as lists.
    :param fieldnames: The field names of this CSV.
    :param mode: The writing mode: Mode.APPEND or Mode.WRITE. By default Mode.WRITE.
    :param encoding: The encoding (it is only used if the parameter file_or_io is a file path).
    """
    with CsvWriter(file_or_io, fieldnames, mode, encoding) as writer:
        writer.write_lists(data)


def dict2csv(file_or_io: Union[str, TextIO, BinaryIO], data: List[Dict],
             fieldnames: Union[List[str], type, object] = None,
             mode: Mode = Mode.WRITE, encoding: str = 'utf-8') -> None:
    """ Write a list of dictionaries in a CSV file.

    :param file_or_io: The file path or an opened stream to use. If it is a file path and it ends in .gz, then
        a compressed file is created using gzip.
    :param data: The list of dictionaries. Each dictionary has to contain as keys the fieldnames and as value
        the row data to store.
    :param fieldnames: The field names of this CSV.
    :param mode: The writing mode: Mode.APPEND or Mode.WRITE. By default Mode.WRITE.
    :param encoding: The encoding (it is only used if the parameter file_or_io is a file path).
    """
    with CsvWriter(file_or_io, fieldnames, mode, encoding) as writer:
        writer.write_dicts(data)


def objects2csv(file_or_io: Union[str, TextIO, BinaryIO], data: List[object],
                fieldnames: Union[List[str], type, object] = None,
                mode: Mode = Mode.WRITE, encoding: str = 'utf-8') -> None:
    """ Write a sequence of objects in a CSV file.

    :param file_or_io: The file path or an opened stream to use. If it is a file path and it ends in .gz, then
        a compressed file is created using gzip.
    :param data: The sequence of objects to write with public attributes or properties.
    :param fieldnames: The field names of this CSV.
    :param mode: The writing mode: Mode.APPEND or Mode.WRITE. By default Mode.WRITE.
    :param encoding: The encoding (it is only used if the parameter file_or_io is a file path).
    """
    with CsvWriter(file_or_io, fieldnames, mode, encoding) as writer:
        writer.write_objects(data)


def list2excel(fname: str, data: List[List[Any]], sheet: Union[str, int] = 0,
               fieldnames: Union[List[str], type, object] = None) -> None:
    """ Write a sequences of lists as a sequence of rows in an Excel file.

    :param fname: The file path to the Excel file.
    :param data: The sequence of rows as lists.
    :param sheet: The sheet to write.
    :param fieldnames: The list of fieldnames. It could be given as a list or a type or object with properties or
    attributes.
    """
    with ExcelWriter(fname, sheet, fieldnames) as writer:
        writer.write_lists(data)


def dict2excel(fname: str, data: List[Dict], sheet: Union[str, int] = 0,
               fieldnames: Union[List[str], type, object] = None) -> None:
    """ Write a list of dictionaries in an Excel file.

    :param fname: The file path to the Excel file.
    :param data: The list of dictionaries. Each dictionary has to contain as keys the fieldnames and as value
        the row data to store.
    :param sheet: The sheet to write.
    :param fieldnames: The list of fieldnames. It could be given as a list or a type or object with properties or
    attributes.
    """
    with ExcelWriter(fname, sheet, fieldnames) as writer:
        writer.write_dicts(data)


def objects2excel(fname: str, data: List[object], sheet: Union[str, int] = 0,
                  fieldnames: Union[List[str], type, object] = None) -> None:
    """ Write a sequence of objects in an Excel file.

    :param fname: The file path to the Excel file.
    :param data: The sequence of objects to write with public attributes or properties.
    :param sheet: The sheet to write.
    :param fieldnames: The list of fieldnames. It could be given as a list or a type or object with properties or
    attributes.
    """
    with ExcelWriter(fname, fieldnames, sheet) as writer:
        writer.write_objects(data)


def csv2list(file_or_io: Union[str, TextIO, BinaryIO], encoding: str = 'utf-8',
             types: Union[List[Type], Dict[str, Type]] = None) -> List[List[Any]]:
    """ Read a CSV file (compressed or not) and return a list of lists with the file rows.

    :param file_or_io: The file path or the file stream.
    :param encoding: The file encoding.
    :param types: The type of each field.
    :return: A List of lists with the file rows. Each column value is stored as list element.
    """
    with CsvReader(file_or_io, ReadMode.LIST, encoding, types) as reader:
        return reader.read_lists()


def csv2dict(file_or_io: Union[str, TextIO, BinaryIO], encoding: str = 'utf-8',
             types: Union[List[Type], Dict[str, Type]] = None) -> List[Dict]:
    """ Read a CSV file (compressed or not) and return a list of dictionaries with the file content..

    :param file_or_io: The file path or the file stream.
    :param encoding: The file encoding.
    :param types: The type of each field.
    :return: A list of dictionaries. Each dictionary represents a row and it contains as keys the column name and
    its value the column value.
    """
    with CsvReader(file_or_io, ReadMode.DICT, encoding, types) as reader:
        return reader.read_rows()


def csv2objects(file_or_io: Union[str, TextIO, BinaryIO], encoding: str = 'utf-8',
                types: Union[List[Type], Dict[str, Type]] = None) -> List[object]:
    """ Read a CSV file (compressed or not) and return a list of objects.

    :param file_or_io: The file path or the file stream.
    :param encoding: The file encoding.
    :param types: The type of each field.
    :return: A list of objects. Each object is a file row with the attributes as column names and its value.
    """
    with CsvReader(file_or_io, ReadMode.OBJECT, encoding, types) as reader:
        return reader.read_objects()


def excel2list(fname: str, sheet: Union[str, int] = 0) -> List[List[Any]]:
    """ Read a Excel file (xlsx or xls) and return a list of lists with the file rows.

    :param fname: The file path to the Excel file..
    :param sheet: The sheet name or the sheet number (starting by 0).
    :return: A List of lists with the file rows. Each column value is stored as list element.
    """
    with ExcelReader(fname, sheet, ReadMode.LIST) as reader:
        return reader.read_lists()


def excel2dict(fname: str, sheet: Union[str, int] = 0) -> List[Dict]:
    """ Read a Excel file (xlsx or xls) and return a list of dictionaries with the file content..

    :param fname: The file path to the Excel file.
    :param sheet: The sheet name or the sheet number (starting by 0).
    :return: A list of dictionaries. Each dictionary represents a row and it contains as keys the column name and
    its value the column value.
    """
    with ExcelReader(fname, sheet, ReadMode.DICT) as reader:
        return reader.read_rows()


def excel2objects(fname: str, sheet: Union[str, int] = 0) -> List[object]:
    """ Read a Excel file (xlsx or xls) and return a list of objects.

    :param fname: The file path to the Excel file.
    :param sheet: The sheet name or the sheet number (starting by 0).
    :return: A list of objects. Each object is a file row with the attributes as column names and its value.
    """
    with ExcelReader(fname, sheet, ReadMode.OBJECT) as reader:
        return reader.read_objects()


def load(fname: str, mode: ReadMode = ReadMode.OBJECT) -> Union[List[object], List[Dict], List[List]]:
    """
    Load an Excel or CSV file and return a list of objects with the file content.
    :param fname: The file path to the file.
    :param mode: The default mode to read the rows. When the reader is iterated,
        it will return objects, dictionaries or lists depending on if the value of this parameter is ReadMode.OBJECT,
        ReadMode.DICTIONARY or ReadMode.LIST, respectively.
    :return: A list of objects. Each object is a file row with the attributes as column names and the attribute values
    as column values.
    """
    if mode == ReadMode.OBJECT:
        return load_objs(fname)
    elif mode == ReadMode.DICT:
        return load_dicts(fname)
    elif mode == ReadMode.LIST:
        return load_lists(fname)


def load_objs(fname: str) -> List[object]:
    """
    Load an Excel or CSV file and return a list of objects with the file content.
    :param fname: The file path to the file.
    :return: A list of objects. Each object is a file row with the attributes as column names and the attribute values
    as column values.
    """
    with open_reader(fname) as reader:
        return [obj for obj in reader]


def load_dicts(fname) -> List[Dict]:
    """
    Load an Excel or CSV file and return a list of dictionaries with the file content.
    :param fname: The file path to the file.
    :return: A list of dictionaries. Each dictionary is a file row with the keys as column names
    and with the dictionary values as column values.
    """
    with open_reader(fname, mode=ReadMode.DICT) as reader:
        return [d for d in reader]


def load_lists(fname) -> List[List]:
    """
    Load an Excel or CSV file and return a list of lists with the file content.
    :param fname: The file path to the file.
    :return: A list of lists. Each list is a file row with the column values.
    """
    with open_reader(fname, mode=ReadMode.LIST) as reader:
        return [lst for lst in reader]


def save(fname: str, objs: List[object]) -> None:
    """
    Save a list of objects as a rows of an Excel or CSV file.
    :param fname: the file path to the file.
    :param objs: The list of objects to write. All objects must have the same attribute or property names.
    :raises ValueError: If the list of objects is empty.
    """
    save_objs(fname, objs)


def save_objs(fname: str, objs: List[object]) -> None:
    """
    Save a list of objects as a rows of an Excel or CSV file.
    :param fname: the file path to the file.
    :param objs: The list of objects to write. All objects must have the same attribute or property names.
    :raises ValueError: If the list of objects is empty.
    """
    if not objs:
        raise ValueError('The list of objects has to contain at least one object to use this method.')
    with open_writer(fname, fieldnames=attributes2list(objs[0])) as writer:
        writer.write_objects(objs)


def save_dicts(fname: str, dicts: List[Dict]) -> None:
    """
    Save a list of dictionaries as a rows of an Excel or CSV file.
    :param fname: the file path to the file.
    :param dicts: The list of dictionaries to write. All the dictionaries must have the same key names.
    :raises ValueError: If the list of dicts is empty.
    """
    if not dicts:
        raise ValueError('The list of dictionaries has to contain at least one dictionary to use this method.')
    with open_writer(fname, fieldnames=dict_keys2list(dicts[0])) as writer:
        writer.write_dicts(dicts)


def save_lists(fname: str, lists: List[List], fieldnames: List[str]) -> None:
    """
    Save a list of lists as a rows of an Excel or CSV file.
    :param fname: the file path to the file.
    :param fieldnames: the list of column names.
    :param lists: The list of lists to write.
    :raises ValueError: If the list is empty.
    """
    if not lists:
        raise ValueError('The list of list has to contain at least one list to use this method.')
    with open_writer(fname, fieldnames=fieldnames) as writer:
        writer.write_lists(lists)


def convert(fr_file: str, to_file) -> None:
    """
    Convert a file into another.
    :param fr_file: The file to copy.
    :param to_file: The target file.
    :raises ValueError: If both files are the same.
    """
    if fr_file == to_file:
        raise ValueError('Both file paths cannot be the same file.')
    with open_reader(fr_file, ReadMode.DICT) as reader:
        with open_writer(to_file, reader.fieldnames) as writer:
            writer.import_reader(reader)



import os
import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


class CleanCommand(setuptools.Command):
    """ Custom clean command to tidy up the project root. """
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        os.system('rm -vrf ./build ./dist ./*.pyc ./*.tgz ./*.egg-info')


class PrepublishCommand(setuptools.Command):
    """  Custom prepublish command. """
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        os.system('python setup.py clean')
        os.system('python setup.py sdist bdist_wheel')


setuptools.setup(
    cmdclass={
        'clean': CleanCommand,
        'prepublish': PrepublishCommand,
    },
    name='filedatasource',
    version='1.1.15',
    url='https://github.com/jmgomezsoriano/filedatasource',
    license='lGPLv3',
    author='José Manuel Gómez Soriano',
    author_email='jmgomez.soriano@gmail.com',
    long_description=long_description,
    long_description_content_type='text/markdown',
    description='Convert several file data sources (like typical CSV or Excel) to python objects in a easy manner.',
    packages=setuptools.find_packages(exclude=["test"]),
    package_dir={'filedatasource': 'filedatasource'},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
    ]
)



from abc import ABC, ABCMeta
from typing import Union, List, Tuple

from filedatasource import Mode
from filedatasource.datafile import DataReader, DataFile, DataWriter, ReadMode


def open_excel(fname: str):
    if fname.lower().endswith('.xlsx'):
        return open_xlsx(fname)
    elif fname.lower().endswith('.xls'):
        return open_xls(fname)
    raise ValueError(f'The file name {fname} has to finish in .xls, or .xlsx to use this method.')


def open_xls(fname: str):
    """  Open an Excel file in the old xls format importing the module.

    :param fname: The path to the xls file.
    :return: The workbook which is a instance of xlrd.book.Book class.
    :raises ModuleNotFoundError: If the module xlrd is not installed.
    """
    try:
        xlrd = __import__('xlrd')
    except ImportError:
        raise ModuleNotFoundError('xlrd is required. Please, install it with:\n\npip install xlrd')
    return xlrd.open_workbook(fname)


def open_xlsx(fname: str):
    """  Open an Excel file in xlsx format importing the module.

    :param fname: The path to the xlsx file.
    :return: The workbook which is a instance of Workbook class.
    :raises ModuleNotFoundError: If the module openpyxl is not installed.
    """
    try:
        openpyxl = __import__('openpyxl')
    except ImportError:
        raise ModuleNotFoundError('openpyxl is required. Please, install it with:\n\npip install openpyxl')
    return openpyxl.load_workbook(fname)


def sheets(fname: str) -> List[str]:
    """ Get the sheet names of the Excel files.

    :param fname: The path to the Excel file.
    :return: A list the seet names of that Excel file.
    """
    doc = open_excel(fname)
    if fname.lower().endswith('.xls'):
        return doc.sheet_names()
    if fname.lower().endswith('.xlsx'):
        return doc.sheetnames


def create_excel(fname: str, sheet: Union[str, int]) -> Tuple['Workbook', 'Worksheet', str]:
    if not fname.lower().endswith('.xlsx') and not fname.lower().endswith('.xls'):
        raise ValueError('The file name has to be the xlsx or xls extensions.')
    if fname.lower().endswith('.xlsx'):
        doc = create_xlsx(fname)
        return doc, doc.add_worksheet(sheet if sheet else 'Sheet'), 'xlsx'
    else:
        doc = create_xls()
        return doc, doc.add_sheet(sheet if sheet else 'Sheet'), 'xls'


def create_xlsx(fname: str) -> 'Workbook':
    """  Create an Excel file in the xlsx format importing the module.

    :param fname: The path to save the xlsx file.
    :return: The workbook which is a instance of Workbook class.
    :raises ModuleNotFoundError: If the module xlsxwriter is not installed.
    """
    try:
        xlsxwriter = __import__('xlsxwriter')
    except ImportError:
        raise ModuleNotFoundError('xlsxwriter is required. Please, install it with:\n\npip install xlsxwriter')
    return xlsxwriter.Workbook(fname)


def create_xls() -> 'Workbook':
    """  Create an Excel file in the old xls format importing the module.

    :return: The workbook which is a instance of xlrd.book.Book class.
    :raises ModuleNotFoundError: If the module xlwt is not installed.
    """
    try:
        xlwt = __import__('xlwt')
    except ImportError:
        raise ModuleNotFoundError('xlwt is required. Please, install it with:\n\npip install xlwt')
    return xlwt.Workbook()


class ExcelData(DataFile, ABC):
    """ Abstract class to define the common attributes of the Excel files. """
    __metaclass__ = ABCMeta

    @property
    def fieldnames(self) -> List[str]:
        """
        :return: The list of fieldnames.
        """
        return self._fieldnames

    @property
    def sheet_name(self) -> str:
        """
        :return:  The sheet name.
        """
        return self.__sheet_name

    @property
    def sheet(self):
        """
        :return: The sheet object.
        """
        return self._sheet

    def __init__(self, fname: str, sheet: Union[str, int] = None) -> None:
        """ Constructor.
        :param fname: The file path to the Excel file.
        :param sheet: The sheet to read/write.
        """
        DataFile.__init__(self, fname)
        self.__sheet_name = sheet if sheet else 0
        self._sheet = None
        self._fieldnames = []


class ExcelReader(ExcelData, DataReader):
    """ The class to read an Excel file easily. """
    def __init__(self, fname: str, sheet: Union[str, int] = 0, mode: ReadMode = ReadMode.OBJECT) -> None:
        """ Constructor.
        :param fname: The file path to the Excel file.
        :param sheet: The sheet to read/write.
        :param mode: The default mode to read the rows. When the reader is iterated,
        it will return objects, dictionaries or lists depending on if the value of this parameter is ReadMode.OBJECT,
        ReadMode.DICTIONARY or ReadMode.LIST, respectively.
        :raises ValueError: If the file name is not a CSV (compressed or not) or Excel (XLSX, XLS) file.
        """
        super(ExcelReader, self).__init__(fname, sheet=sheet)
        DataReader.__init__(self, fname, mode=mode)
        if fname.lower().endswith('.xlsx'):
            self.__doc = open_xlsx(fname)
            self._sheet = self.__doc[sheet] if isinstance(sheet, str) else self.__doc[self.__doc.sheetnames[sheet]]
            self._fieldnames = [cell.value for cell in next(self.sheet.rows)]
            self.__type = 'xlsx'
        elif fname.lower().endswith('.xls'):
            doc = open_xls(fname)
            self.__doc = doc
            self._sheet = doc.sheet_by_name(sheet) if isinstance(sheet, str) else doc.sheet_by_index(sheet)
            self._fieldnames = [cell.value for cell in self.sheet.row(0)]
            self.__type = 'xls'
        else:
            raise ValueError(f'The file name {fname} has to end in .xls or .xlsx.')

        self.__row = 1

    def read_row(self) -> dict:
        """ Read a row of the Excel file as a dict.

        :return: A dictionary where the keys are the fieldnames, and their values the row values.
        """
        sheet = self.sheet
        if self.__row < (sheet.max_row if self.__type == 'xlsx' else sheet.nrows):
            self.__row += 1
            if self.__type == 'xlsx':
                return self.__read_xlsx_row(sheet)
            else:
                return self.__read_xls_row(sheet)
        raise StopIteration()

    def close(self) -> None:
        """ Nothing to do. """
        pass

    def __read_xlsx_row(self, sheet) -> dict:
        """ Write a row using openpyxl module.
        :param sheet: The sheet to write the row.
        :return: A dict with the fieldnames as keys.
        """
        return {self.fieldnames[i]: sheet.cell(row=self.__row, column=i + 1).value for i in range(sheet.max_column)}

    def __read_xls_row(self, sheet) -> dict:
        """ Write a row using xlrd module.
        :param sheet: The sheet to write the row.
        :return: A dict with the fieldnames as keys.
        """
        return {self.fieldnames[i]: cell.value for i, cell in enumerate(sheet.row(self.__row - 1))}

    def __len__(self) -> int:
        """
        :return: The number of rows.
        """
        return (self.sheet.max_row if self.__type == 'xlsx' else self.sheet.nrows) - 1


class ExcelWriter(ExcelData, DataWriter):
    """ The class to create an Excel file easily """
    def __init__(self, fname: str, sheet: Union[str, int] = 0, fieldnames: Union[List[str], type, object] = None,
                 mode: Mode = Mode.WRITE):
        """ Constructor.
        :param fname: The file path to the Excel file.
        :param sheet: The sheet to read/write.
        :param fieldnames: The list of fieldnames. It could be given as a list or a type or object with properties or
        attributes.
        :param mode: The writing mode: Mode.APPEND or Mode.WRITE. By default Mode.WRITE.
        """
        DataWriter.__init__(self, fname, mode)
        ExcelData.__init__(self, fname, sheet=sheet)
        self.__num_row = 0
        if mode == Mode.WRITE:
            if not fieldnames:
                raise ValueError('The fieldnames parameters must contain a least a field name in write mode.')
            self._fieldnames = self._parse_fieldnames(fieldnames)
            self._doc, self._sheet, self.__type = create_excel(fname, sheet)
            self.write_list(self.fieldnames)
        elif mode == Mode.APPEND:
            with ExcelReader(fname, sheet, ReadMode.DICT) as reader:
                self._fieldnames = self._parse_fieldnames(fieldnames) if fieldnames else reader.fieldnames
                self._doc, self._sheet, self.__type = create_excel(fname, sheet)
                self.write_list(self.fieldnames)
                self.import_reader(reader)

    def write_row(self, **row) -> None:
        """ Append a row to the sheet.

        :param row: The row to append.
        """
        sheet_row = self.sheet.row(self.__num_row) if self.__type == 'xls' else self.sheet
        for i, field in enumerate(self.fieldnames):
            if field in row:
                if self.__type == 'xls':
                    sheet_row.write(i, row[field])
                else:
                    sheet_row.write(self.__num_row, i, row[field])
        self.__num_row += 1

    def close(self) -> None:
        """ Close saving the file. """
        if self.__type == 'xls':
            self._doc.save(self.file_name)
        else:
            self._doc.close()

    def __len__(self) -> int:
        """
        :return: The number of rows.
        """
        return self.__num_row - 1
